var initPlayer = (function () {
    var defaultFailureTimeoutMillis = 2000;
    var hlsPingIntervalMillis = 1000;
    var hlsPingIntervalTimes = 15;
    
    function setupDash (options)
    {
        var url = options.root + "dash/" + options.stream.truncatedID + ".mpd";
        var player = dashjs.MediaPlayer().create();
        player.initialize(options.player, url, true);
        return jQuery.Deferred().resolve();
    }
    
    function attachRtmp (options)
    {
        var player = options.player;
        var source = document.createElement("source");
        source.type = "rtmp/flv";
        source.src = "rtmp://" + options.host + "/preview/" + options.stream.truncatedID;
        player.appendChild(source);
        return jQuery.Deferred().resolve();
    }
    
    function attachCws (options)
    {
        var player = options.player;
        var source = document.createElement("source");
        source.type = 'video/webm;codecs="vorbis,vp8"';
        source.src = options.root + "stream-" + options.port + "/" + options.stream.truncatedID + "/" + options.stream.width + "x" + options.stream.height + "/" + Math.round(1000000 * Math.random()) + ".webm";
        player.appendChild(source);
        return jQuery.Deferred().resolve();
    }
    
    function attachHls (options)
    {
        /*
         * Attach HLS source when it becomes available.
         * If the stream is just starting, HLS might not be available
         * just yet, so poll until it is.
         */
        var hlsURL = options.root + "hls/" + options.stream.truncatedID + ".m3u8",
            deferred = jQuery.Deferred(),
            setupFunction = function () {
                var player = options.player;
                var source = document.createElement("source");
                source.type = "application/x-mpegURL";
                source.src = hlsURL;
                player.appendChild(source);
                deferred.resolve();
            },
            testFunction = function (i, numChecks, waitBetweenMillis) {
                if (i >= numChecks)
                {
                    deferred.resolve();
                    return;
                }

                jQuery.get(hlsURL).then(setupFunction).fail(function () {
                    setTimeout(function () {
                        testFunction(i + 1, numChecks, waitBetweenMillis);
                    }, waitBetweenMillis);
                });
            };

        testFunction(0, hlsPingIntervalTimes, hlsPingIntervalMillis);
        return deferred.promise();
    }
    
    function debugEvents (player)
    {
        var evts = ["loadstart", "play", "pause", "playing", "ended",
                    "volumechange", "durationchange", "error", "timeupdate",
                    "progress", "enterFullScreen", "exitFullScreen"];

        for (var i = 0; i < evts.length; i++)
        {
            (function (evt) {
                player.on(evt, function () {
                    console.log(new Date(), evt, arguments);
                });
            })(evts[i]);
        }
    }
    
    function setupVideoJs (options)
    {
        var deferred = jQuery.Deferred(),
            player = options.player;

        player.controls = false;
        player.width = options.stream.width;
        player.height = options.stream.height;

        videojs(player, jQuery.extend({}, {
            autoplay: true,
            controls: false,
            width: options.stream.width,
            height: options.stream.height,
            techOrder: ["flash", "html5"]
        }, options.options), function () {
            var player = this;
            deferred.resolve(player);
        });

        return deferred.promise();
    }
    
    /*
     * Wait on a video.js event to fire so we know
     * the video is loaded. If it doesn't fire in a certain
     * timeout, consider it failed so we can handle it
     * earlier. video.js's flash object likes to wait 3 - 4 minutes
     * before emitting an error.
     */
    function handleLoadFailure (player, options)
    {
        options = options || {};
        var deferred = jQuery.Deferred(),
            hasLoadedSuccessfully = false,
            events = options.events || ["playing", "loadstart"],
            timeout = options.timeout || defaultFailureTimeoutMillis;
        
        for (var i = 0; i < events.length; i++)
        {
            player.on(events[i], function () {
                hasLoadedSuccessfully = true;
                deferred.resolve(player);
            });            
        }
        
        setTimeout(function () {
            if (hasLoadedSuccessfully)
            {
                deferred.resolve(player);
            }
            else
            {
                deferred.reject(player);
            }
        }, timeout);

        return deferred.promise();
    }
    
    function startPinger (group, stream, client, pingIntervalMillis)
    {
        setInterval(function () {
            jQuery.post("?clientId=" + client.id);
        }, pingIntervalMillis);
    }
    
    function init (options)
    {
        /*
         * Generate video tag here. VideoJS will replace it with its own
         * content when we init it, so we need to recreate it if we
         * need to refresh VideoJS.
         */
        var tagId = options.id || "videoPlayer",
            enableRtmp = options.rtmp ? options.rtmp.enabled : true,
            enableWebm = options.cws ? options.cws.enabled : true,
            root = options.root,
            player = jQuery("<video>").attr("id", tagId).prependTo("body"),
            stream = options.stream,
            deferred = jQuery.Deferred();
        
        deferred.resolve(true);
        
        if (enableRtmp && stream.formats.indexOf("rtmp") > -1)
        {
            var rtmpHost = options.rtmp.host || options.rtmp.hosts[0];
            
            if (rtmpHost)
            {
                console.log("trying with rtmp host", rtmpHost);
                deferred = deferred.pipe(function () {
                    return attachRtmp({
                        host: options.rtmp.host || options.rtmp.hosts[0],
                        player: player[0],
                        stream: stream
                    });
                });

                /*deferred = deferred.pipe(function () {
                    return attachHls({
                        player: player[0],
                        stream: stream,
                        root: root                
                    });
                });*/
            }
        }

        if (enableWebm && stream.formats.indexOf("webm") > -1)
        {
            console.log("using CWS port", options.cws.port);
            deferred = deferred.pipe(function () {
                return attachCws({
                    player: player[0],
                    stream: stream,
                    root: root,
                    port: options.cws.port
                });
            });;
        }
        
        deferred.pipe(function () {
            var startTimestampMillis = new Date().getTime();
            setupVideoJs({
                player: player[0],
                stream: stream,
                options: options.videojs
            }).then(function (player) {
                handleLoadFailure(player).done(function () {
                    console.log("load success", player.techName_, player.currentType_, new Date().getTime() - startTimestampMillis, "ms");
                    startPinger(options.group, options.stream, options.client, options.pingIntervalMillis);
                }).fail(function () {
                    /*
                     * If RTMP (Flash) player fails to load with the given host,
                     * eliminate it from the possible hosts for RTMP and try
                     * the next one. If none of the possible RTMP hosts can be
                     * loaded, force the WebM fallback.
                     * 
                     * In order to do this, we need to dispose the VideoJS player 
                     * and re-instantiate it with the new RTMP settings.
                     */
                    console.log("load fail", player.techName_, player.currentType_, new Date().getTime() - startTimestampMillis, "ms");
                    player.dispose();
                    init(jQuery.extend({}, options, {
                        rtmp: jQuery.extend({}, options.rtmp, {
                            hosts: options.rtmp.hosts.slice(1)
                        })
                    }));
                });
            });
        });
    }
    
    return init;
})();